package Win32::file;

use Exporter;
use DynaLoader;
use Carp;
use Fcntl;

@ISA=qw(Exporter DynaLoader);
@EX_DEFS=qw(_LK_LOCK _LK_RLCK _LK_NBLCK _LK_NBRLCK _LK_UNLCK
             LK_LOCK  LK_RLCK  LK_NBLCK  LK_NBRLCK  LK_UNLCK
            SEEK_SET SEEK_CUR SEEK_END
            S_IRWXU S_IRUSR S_IWUSR S_IXUSR
            S_IRWXG S_IRGRP S_IWGRP S_IXGRP
            S_IRWXO S_IROTH S_IWOTH S_IXOTH
            S_IREAD S_IWRITE S_IEXEC
            );

@EX_FUNCS=qw(flock);
@EX_WIN32=qw(GetWindowsDirectory GetSystemDirectory GetCurrentDirectory);
@EX_XFUNCS=qw(_locking _open _close _write _read _lseek _tell _setmode _filelength);

@EXPORT=(@EX_DEFS, @EX_FUNCS, @EX_WIN32);
%EXPORT_TAGS=(  FUNCTIONS=>[@EX_FUNCS],
                DEFINES=>[@EX_DEFS],
                XFUNCTIONS=>[@EX_XFUNCS],
                WIN32=>[@EX_WIN32],
            );

bootstrap Win32::file;

#######################################################################
#### PACKAGE GLOBALS
use strict;

$Win32::file::VERSION='0.01';
$Win32::file::PACKAGE='Win32::CommDlg';

use vars qw( %_myFiles );

#######################################################################
#### PROTOTYPES
sub flock (*$);

sub _locking ($$$);
sub _open ($$$);
sub _close ($);
sub _read ($$$);
sub _write ($$$);
sub _setmode($$);
sub _lseek($$$);
sub _tell($);
sub _filelength($);


sub _niceHndl(*);
#######################################################################
sub _niceHndl (*)
{
    my($hndl)=@_;
    return $hndl if(ref $hndl);

    my($package) = caller(1);               # handle in the right package
    $hndl = $hndl=~/[':]/ ? $hndl : "*$package\:\:$hndl";
    $hndl=~s/'/::/g;
    return $hndl;
}

#######################################################################
### AUTOLOAD
# This AUTOLOAD is used to 'autoload' constants from the constant()
# XS function.  If a constant is not found then control is passed
# to the AUTOLOAD in AutoLoader.
sub AUTOLOAD {
    my($constname);
    ($constname = $Win32::file::AUTOLOAD) =~ s/.*:://;
    #reset $! to zero to reset any current errors.
    $!=0;

#    my $val = constant($constname, @_ ? $_[0] : 0);
    my $val = constant($constname, 0);

    if ($! != 0)
    {

        # [dada] This results in an ugly Autoloader error
        #if ($! =~ /Invalid/) {
        #    $AutoLoader::AUTOLOAD = $AUTOLOAD;
        #    goto &AutoLoader::AUTOLOAD;
        #} else {
        # [dada] ... I prefer this one :)

            my($pack,$file,$line) = caller;
            die "Win32::file::$constname is not defined, used at $file line $line.\n";
            $pack=$pack if($^W);
        #}
    }
    # printf STDERR "- $constname = 0x%X\n", $val;
    eval "sub $Win32::file::AUTOLOAD { $val }";
    goto &$Win32::file::AUTOLOAD;
}


#######################################################################
sub flock
{
    my($file, $lock)=@_;
    $file=_niceHndl($file);
    my $hndl=fileno($file);

    # Because we can't share file handles with DLLs
    # we create a lock file in the windows directory.
    # It can't go in the System directory because we probably
    # don't have write permission there.

    my $name=GetWindowsDirectory();
    $name.='\\' unless($name=~/\\$/);   # windows-dir can sometimes have a
    $name.="perl.$$.$hndl.lock";        # trailing \  ie c:\ (not likely)

    # convert from unix-style defines to ms-dos defines

    my $true_lock=0;
    if($lock & 1)                       # LOCK_SH (sharing)
    {   $true_lock |= 0; }              # What is the use of this then?
    if($lock & 2)                       # LOCK_EX (exclusive)
    {   $true_lock |= &_LK_LOCK; }
    if($lock & 4)                       # LOCK_NB (non-bfile)
    {   $true_lock |= &_LK_NBLCK; }
    if($lock & 8)                       # LOCK_UN (unlock)
    {
        if($_myFiles{$name})            # if we'd already locked the file
        {
            _close  $_myFiles{$name};   # close the file
            delete $_myFiles{$name};    # kill our memory
            unlink $name;               # kill the file
        } else
        {}                              # we never locked it, but we'll let it
                                        # ride for now
        return 1;                       # all cool
    }

    # create the file (or kill one that was there before)
    my $handle=_open($name, &O_CREAT|&O_TRUNC|&O_WRONLY,
                      &S_IRUSR|&S_IWUSR|&S_IRGRP|&S_IWGRP|&S_IROTH|&S_IWOTH);

    # now, this should happen, but it will, so we try to be
    # inteligent about it
    if(-1==$handle)
    {
        carp "Unable to create $name ($!), file not locked."; # if($^W);
        $!=&EDEADLOCK;
        return 0;
    }

    # fill it with our PID.  Why?  Well... because. -:)
    my $ret="$$";
    my $size=_write($handle, $ret, length($ret));

    # now we grab the lock
    my $err;
    while($ret!=0)
    {
        $ret=_locking($handle, $true_lock, $size);
        $err=$!;                        # save this for later
        # print "ret=$ret: $!\n";
        last if($lock&4);               # non-blocking
        last unless($! == &EDEADLOCK);  # this means someone else has the lock
                                        # so we loop and wait
                                        # all other things are errors
    }

    if($ret==0)                         # We got the lock!!!
    {
        $_myFiles{$name}=$handle;       # so we must remember what's going on
        return 1;
    }

    _close($handle);                    # clean up our mess
    unlink($name)
        or warn "Unable to unlink $name: $!\n";
    $!=$err;                            # this is the real errro that the
    return 0;                           # caller want's to see
}

#######################################################################
# tidy up our mess
END
{
    my $name;
    foreach $name (keys %_myFiles)
    {
        _close  $_myFiles{$name};   # close the file
        delete $_myFiles{$name};    # kill our memory
        unlink $name                # kill the file
            or warn "Unable to unlink $name: $!\n";
    }
}

1;
